from typing import Union

from . import RequestMsg, ReplyMsg, Message, SimpleMessage
from hedgehog.protocol.errors import InvalidCommandError
from hedgehog.protocol.proto import io_pb2
from hedgehog.protocol.proto.io_pb2 import INPUT_FLOATING, INPUT_PULLUP, INPUT_PULLDOWN
from hedgehog.protocol.proto.io_pb2 import OUTPUT_OFF, OUTPUT_ON
from hedgehog.protocol.proto.io_pb2 import OUTPUT, PULLUP, PULLDOWN, LEVEL
from hedgehog.protocol.proto.subscription_pb2 import Subscription
from hedgehog.utils import protobuf


def _check_flags(flags: int) -> None:
    if flags & OUTPUT and flags & (PULLUP | PULLDOWN):
        raise InvalidCommandError("only input ports can be set to pullup or pulldown")
    if not flags & OUTPUT and flags & LEVEL:
        raise InvalidCommandError("only output ports can be set to on")
    if flags & PULLUP and flags & PULLDOWN:
        raise InvalidCommandError("pullup and pulldown are mutually exclusive")


@RequestMsg.message(io_pb2.IOAction, 'io_action')
class Action(SimpleMessage):
    def __init__(self, port: int, flags: int) -> None:
        _check_flags(flags)
        self.port = port
        self.flags = flags

    @property
    def output(self) -> bool:
        return (self.flags & OUTPUT) != 0

    @property
    def pullup(self) -> bool:
        return (self.flags & PULLUP) != 0

    @property
    def pulldown(self) -> bool:
        return (self.flags & PULLDOWN) != 0

    @property
    def level(self) -> bool:
        return (self.flags & LEVEL) != 0

    @classmethod
    def _parse(cls, msg: io_pb2.IOAction) -> 'Action':
        port = msg.port
        flags = msg.flags
        return cls(port, flags)

    def _serialize(self, msg: io_pb2.IOAction) -> None:
        msg.port = self.port
        msg.flags = self.flags


@protobuf.message(io_pb2.IOCommandMessage, 'io_command_message', fields=('port',))
class CommandRequest(Message):
    def __init__(self, port: int) -> None:
        self.port = port

    def _serialize(self, msg: io_pb2.IOCommandMessage) -> None:
        msg.port = self.port


@protobuf.message(io_pb2.IOCommandMessage, 'io_command_message', fields=('port', 'subscription'))
class CommandSubscribe(Message):
    def __init__(self, port: int, subscription: Subscription) -> None:
        self.port = port
        self.subscription = subscription

    def _serialize(self, msg: io_pb2.IOCommandMessage) -> None:
        msg.port = self.port
        msg.subscription.CopyFrom(self.subscription)


@RequestMsg.parser('io_command_message')
def _parse_command_request(msg: io_pb2.IOCommandMessage) -> Union[CommandRequest, CommandSubscribe]:
    port = msg.port
    subscription = msg.subscription if msg.HasField('subscription') else None
    if subscription is None:
        return CommandRequest(port)
    else:
        return CommandSubscribe(port, subscription)


@protobuf.message(io_pb2.IOCommandMessage, 'io_command_message', fields=('port', 'flags'))
class CommandReply(Message):
    def __init__(self, port: int, flags: int) -> None:
        _check_flags(flags)
        self.port = port
        self.flags = flags

    @property
    def output(self) -> bool:
        return (self.flags & OUTPUT) != 0

    @property
    def pullup(self) -> bool:
        return (self.flags & PULLUP) != 0

    @property
    def pulldown(self) -> bool:
        return (self.flags & PULLDOWN) != 0

    @property
    def level(self) -> bool:
        return (self.flags & LEVEL) != 0

    def _serialize(self, msg: io_pb2.IOCommandMessage) -> None:
        msg.port = self.port
        msg.flags = self.flags


@protobuf.message(io_pb2.IOCommandMessage, 'io_command_message')
class CommandUpdate(Message):
    is_async = True

    def __init__(self, port: int, flags: int, subscription: Subscription) -> None:
        _check_flags(flags)
        self.port = port
        self.flags = flags
        self.subscription = subscription

    @property
    def output(self) -> bool:
        return (self.flags & OUTPUT) != 0

    @property
    def pullup(self) -> bool:
        return (self.flags & PULLUP) != 0

    @property
    def pulldown(self) -> bool:
        return (self.flags & PULLDOWN) != 0

    @property
    def level(self) -> bool:
        return (self.flags & LEVEL) != 0

    def _serialize(self, msg: io_pb2.IOCommandMessage) -> None:
        msg.port = self.port
        msg.flags = self.flags
        msg.subscription.CopyFrom(self.subscription)


@ReplyMsg.parser('io_command_message')
def _parse_command_reply(msg: io_pb2.IOCommandMessage) -> Union[CommandReply, CommandUpdate]:
    port = msg.port
    flags = msg.flags
    subscription = msg.subscription if msg.HasField('subscription') else None
    if subscription is None:
        return CommandReply(port, flags)
    else:
        return CommandUpdate(port, flags, subscription)



from hedgehog.protocol import CommSide
from hedgehog.utils.zmq.async_socket import Socket
from . import RawMessage, Message, RawPayload, Payload, \
    Header, RawMsgs, Msgs, RawMsg, Msg

from .sockets import raw_to_delimited, to_delimited, raw_from_delimited, from_delimited


class DealerRouterMixin(object):
    """
    A mixin for async ZMQ dealer & router sockets used to send delimited & Hedgehog-encoded messages.

    This mixin defines methods to send/receive single/multipart binary/hedgehog messages on dealer & router sockets.
    For example, `send_msg` send a single hedgehog message, while `recv_msgs_raw` receives a multipart binary message.
    All these methods use a header (one or more binary frames) followed by a delimiter (one empty frame). `send` methods
    accept a header parameter before the payload, `recv` methods return the header and payload as a tuple.
    """
    async def send_msg(self, header: Header, msg: Message):
        await self.send_msgs(header, [msg])

    async def recv_msg(self) -> Msg:
        header, [msg] = await self.recv_msgs()
        return header, msg

    async def send_msgs(self, header: Header, msgs: Payload):
        await self.send_msgs_raw(header, [self.side.serialize(msg) for msg in msgs])

    async def recv_msgs(self) -> Msgs:
        header, msgs_raw = await self.recv_msgs_raw()
        return header, tuple(self.side.parse(msg_raw) for msg_raw in msgs_raw)

    async def send_msg_raw(self, header: Header, msg_raw: RawMessage):
        await self.send_msgs_raw(header, [msg_raw])

    async def recv_msg_raw(self) -> RawMsg:
        header, [msg_raw] = await self.recv_msgs_raw()
        return header, msg_raw

    async def send_msgs_raw(self, header: Header, msgs_raw: RawPayload):
        await self.send_multipart(raw_to_delimited(header, msgs_raw))

    async def recv_msgs_raw(self) -> RawMsgs:
        return raw_from_delimited(await self.recv_multipart())


class DealerRouterSocket(DealerRouterMixin, Socket):
    side = None

    def __init__(self, *args, side: CommSide, **kwargs) -> None:
        super(DealerRouterSocket, self).__init__(*args, **kwargs)
        self.side = side


class ReqMixin(object):
    """
    A mixin for async ZMQ req sockets used to send Hedgehog-encoded messages.

    This mixin defines methods to send/receive single/multipart binary/hedgehog messages on req sockets.
    For example, `send_msg` send a single hedgehog message, while `recv_msgs_raw` receives a multipart binary message.
    All these methods use a delimiter (one empty frame), implicitly added by the req socket.
    """

    async def send_msg(self, msg: Message):
        await self.send_msgs([msg])

    async def recv_msg(self) -> Message:
        [msg] = await self.recv_msgs()
        return msg

    async def send_msgs(self, msgs: Payload):
        await self.send_msgs_raw([self.side.serialize(msg) for msg in msgs])

    async def recv_msgs(self) -> Payload:
        return tuple(self.side.parse(msg_raw) for msg_raw in await self.recv_msgs_raw())

    async def send_msg_raw(self, msg_raw: RawMessage):
        await self.send_msgs_raw([msg_raw])

    async def recv_msg_raw(self) -> RawMessage:
        [msg_raw] = await self.recv_msgs_raw()
        return msg_raw

    async def send_msgs_raw(self, msgs_raw: RawPayload):
        await self.send_multipart(msgs_raw)

    async def recv_msgs_raw(self) -> RawPayload:
        return await self.recv_multipart()


class ReqSocket(ReqMixin, Socket):
    side = None

    def __init__(self, *args, side: CommSide, **kwargs) -> None:
        super(ReqSocket, self).__init__(*args, **kwargs)
        self.side = side



"""
Errors that may be caused by Hedgehog commands.

Every error corresponds to one acknowledge code from ack.proto; the `OK` code naturally has no corresponding error.
"""

from typing import Dict, Type

from .proto.ack_pb2 import UNKNOWN_COMMAND, INVALID_COMMAND, UNSUPPORTED_COMMAND, FAILED_COMMAND


class HedgehogCommandError(Exception):
    """Superclass of all errors caused by Hedgehog commands."""
    code = None  # type: int
    """Class property containing the acknowledgement code"""

    def to_message(self):
        """
        Creates an error Acknowledgement message.
        The message's code and message are taken from this exception.

        :return: the message representing this exception
        """
        from .messages import ack
        return ack.Acknowledgement(self.code, self.args[0] if len(self.args) > 0 else '')


class UnknownCommandError(HedgehogCommandError):
    code = UNKNOWN_COMMAND


class InvalidCommandError(HedgehogCommandError):
    code = INVALID_COMMAND


class UnsupportedCommandError(HedgehogCommandError):
    code = UNSUPPORTED_COMMAND


class FailedCommandError(HedgehogCommandError):
    code = FAILED_COMMAND


class EmergencyShutdown(FailedCommandError):
    pass


_errors = {
    UNKNOWN_COMMAND: UnknownCommandError,
    INVALID_COMMAND: InvalidCommandError,
    UNSUPPORTED_COMMAND: UnsupportedCommandError,
    FAILED_COMMAND: FailedCommandError,
}  # type: Dict[int, Type[HedgehogCommandError]]


def error(code: int, *args, **kwargs) -> HedgehogCommandError:
    """
    Creates an error from the given code, and args and kwargs.

    :param code: The acknowledgement code
    :param args: Exception args
    :param kwargs: Exception kwargs
    :return: the error for the given acknowledgement code
    """
    # TODO add proper error code
    if code == FAILED_COMMAND and len(args) >= 1 and args[0] == "Emergency Shutdown activated":
        return EmergencyShutdown(*args, **kwargs)
    return _errors[code](*args, **kwargs)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: hedgehog/protocol/proto/hedgehog.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from hedgehog.protocol.proto import ack_pb2 as hedgehog_dot_protocol_dot_proto_dot_ack__pb2
from hedgehog.protocol.proto import io_pb2 as hedgehog_dot_protocol_dot_proto_dot_io__pb2
from hedgehog.protocol.proto import motor_pb2 as hedgehog_dot_protocol_dot_proto_dot_motor__pb2
from hedgehog.protocol.proto import servo_pb2 as hedgehog_dot_protocol_dot_proto_dot_servo__pb2
from hedgehog.protocol.proto import process_pb2 as hedgehog_dot_protocol_dot_proto_dot_process__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='hedgehog/protocol/proto/hedgehog.proto',
  package='hedgehog.protocol.proto',
  syntax='proto3',
  serialized_pb=_b('\n&hedgehog/protocol/proto/hedgehog.proto\x12\x17hedgehog.protocol.proto\x1a!hedgehog/protocol/proto/ack.proto\x1a hedgehog/protocol/proto/io.proto\x1a#hedgehog/protocol/proto/motor.proto\x1a#hedgehog/protocol/proto/servo.proto\x1a%hedgehog/protocol/proto/process.proto\"\xae\t\n\x0fHedgehogMessage\x12\x43\n\x0f\x61\x63knowledgement\x18\x01 \x01(\x0b\x32(.hedgehog.protocol.proto.AcknowledgementH\x00\x12\x36\n\tio_action\x18\x02 \x01(\x0b\x32!.hedgehog.protocol.proto.IOActionH\x00\x12G\n\x12io_command_message\x18\x10 \x01(\x0b\x32).hedgehog.protocol.proto.IOCommandMessageH\x00\x12@\n\x0e\x61nalog_message\x18\x03 \x01(\x0b\x32&.hedgehog.protocol.proto.AnalogMessageH\x00\x12\x42\n\x0f\x64igital_message\x18\x04 \x01(\x0b\x32\'.hedgehog.protocol.proto.DigitalMessageH\x00\x12<\n\x0cmotor_action\x18\x05 \x01(\x0b\x32$.hedgehog.protocol.proto.MotorActionH\x00\x12M\n\x15motor_command_message\x18\x11 \x01(\x0b\x32,.hedgehog.protocol.proto.MotorCommandMessageH\x00\x12I\n\x13motor_state_message\x18\x06 \x01(\x0b\x32*.hedgehog.protocol.proto.MotorStateMessageH\x00\x12T\n\x19motor_set_position_action\x18\x12 \x01(\x0b\x32/.hedgehog.protocol.proto.MotorSetPositionActionH\x00\x12<\n\x0cservo_action\x18\x07 \x01(\x0b\x32$.hedgehog.protocol.proto.ServoActionH\x00\x12M\n\x15servo_command_message\x18\x13 \x01(\x0b\x32,.hedgehog.protocol.proto.ServoCommandMessageH\x00\x12O\n\x16process_execute_action\x18\x14 \x01(\x0b\x32-.hedgehog.protocol.proto.ProcessExecuteActionH\x00\x12M\n\x15process_execute_reply\x18\x15 \x01(\x0b\x32,.hedgehog.protocol.proto.ProcessExecuteReplyH\x00\x12O\n\x16process_stream_message\x18\x08 \x01(\x0b\x32-.hedgehog.protocol.proto.ProcessStreamMessageH\x00\x12M\n\x15process_signal_action\x18\x16 \x01(\x0b\x32,.hedgehog.protocol.proto.ProcessSignalActionH\x00\x12I\n\x13process_exit_update\x18\x17 \x01(\x0b\x32*.hedgehog.protocol.proto.ProcessExitUpdateH\x00\x42\t\n\x07payloadB,\n\x1f\x61t.pria.hedgehog.protocol.protoB\tHedgehogPb\x06proto3')
  ,
  dependencies=[hedgehog_dot_protocol_dot_proto_dot_ack__pb2.DESCRIPTOR,hedgehog_dot_protocol_dot_proto_dot_io__pb2.DESCRIPTOR,hedgehog_dot_protocol_dot_proto_dot_motor__pb2.DESCRIPTOR,hedgehog_dot_protocol_dot_proto_dot_servo__pb2.DESCRIPTOR,hedgehog_dot_protocol_dot_proto_dot_process__pb2.DESCRIPTOR,])
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_HEDGEHOGMESSAGE = _descriptor.Descriptor(
  name='HedgehogMessage',
  full_name='hedgehog.protocol.proto.HedgehogMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='acknowledgement', full_name='hedgehog.protocol.proto.HedgehogMessage.acknowledgement', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='io_action', full_name='hedgehog.protocol.proto.HedgehogMessage.io_action', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='io_command_message', full_name='hedgehog.protocol.proto.HedgehogMessage.io_command_message', index=2,
      number=16, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='analog_message', full_name='hedgehog.protocol.proto.HedgehogMessage.analog_message', index=3,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='digital_message', full_name='hedgehog.protocol.proto.HedgehogMessage.digital_message', index=4,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='motor_action', full_name='hedgehog.protocol.proto.HedgehogMessage.motor_action', index=5,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='motor_command_message', full_name='hedgehog.protocol.proto.HedgehogMessage.motor_command_message', index=6,
      number=17, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='motor_state_message', full_name='hedgehog.protocol.proto.HedgehogMessage.motor_state_message', index=7,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='motor_set_position_action', full_name='hedgehog.protocol.proto.HedgehogMessage.motor_set_position_action', index=8,
      number=18, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='servo_action', full_name='hedgehog.protocol.proto.HedgehogMessage.servo_action', index=9,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='servo_command_message', full_name='hedgehog.protocol.proto.HedgehogMessage.servo_command_message', index=10,
      number=19, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='process_execute_action', full_name='hedgehog.protocol.proto.HedgehogMessage.process_execute_action', index=11,
      number=20, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='process_execute_reply', full_name='hedgehog.protocol.proto.HedgehogMessage.process_execute_reply', index=12,
      number=21, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='process_stream_message', full_name='hedgehog.protocol.proto.HedgehogMessage.process_stream_message', index=13,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='process_signal_action', full_name='hedgehog.protocol.proto.HedgehogMessage.process_signal_action', index=14,
      number=22, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='process_exit_update', full_name='hedgehog.protocol.proto.HedgehogMessage.process_exit_update', index=15,
      number=23, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='payload', full_name='hedgehog.protocol.proto.HedgehogMessage.payload',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=250,
  serialized_end=1448,
)

_HEDGEHOGMESSAGE.fields_by_name['acknowledgement'].message_type = hedgehog_dot_protocol_dot_proto_dot_ack__pb2._ACKNOWLEDGEMENT
_HEDGEHOGMESSAGE.fields_by_name['io_action'].message_type = hedgehog_dot_protocol_dot_proto_dot_io__pb2._IOACTION
_HEDGEHOGMESSAGE.fields_by_name['io_command_message'].message_type = hedgehog_dot_protocol_dot_proto_dot_io__pb2._IOCOMMANDMESSAGE
_HEDGEHOGMESSAGE.fields_by_name['analog_message'].message_type = hedgehog_dot_protocol_dot_proto_dot_io__pb2._ANALOGMESSAGE
_HEDGEHOGMESSAGE.fields_by_name['digital_message'].message_type = hedgehog_dot_protocol_dot_proto_dot_io__pb2._DIGITALMESSAGE
_HEDGEHOGMESSAGE.fields_by_name['motor_action'].message_type = hedgehog_dot_protocol_dot_proto_dot_motor__pb2._MOTORACTION
_HEDGEHOGMESSAGE.fields_by_name['motor_command_message'].message_type = hedgehog_dot_protocol_dot_proto_dot_motor__pb2._MOTORCOMMANDMESSAGE
_HEDGEHOGMESSAGE.fields_by_name['motor_state_message'].message_type = hedgehog_dot_protocol_dot_proto_dot_motor__pb2._MOTORSTATEMESSAGE
_HEDGEHOGMESSAGE.fields_by_name['motor_set_position_action'].message_type = hedgehog_dot_protocol_dot_proto_dot_motor__pb2._MOTORSETPOSITIONACTION
_HEDGEHOGMESSAGE.fields_by_name['servo_action'].message_type = hedgehog_dot_protocol_dot_proto_dot_servo__pb2._SERVOACTION
_HEDGEHOGMESSAGE.fields_by_name['servo_command_message'].message_type = hedgehog_dot_protocol_dot_proto_dot_servo__pb2._SERVOCOMMANDMESSAGE
_HEDGEHOGMESSAGE.fields_by_name['process_execute_action'].message_type = hedgehog_dot_protocol_dot_proto_dot_process__pb2._PROCESSEXECUTEACTION
_HEDGEHOGMESSAGE.fields_by_name['process_execute_reply'].message_type = hedgehog_dot_protocol_dot_proto_dot_process__pb2._PROCESSEXECUTEREPLY
_HEDGEHOGMESSAGE.fields_by_name['process_stream_message'].message_type = hedgehog_dot_protocol_dot_proto_dot_process__pb2._PROCESSSTREAMMESSAGE
_HEDGEHOGMESSAGE.fields_by_name['process_signal_action'].message_type = hedgehog_dot_protocol_dot_proto_dot_process__pb2._PROCESSSIGNALACTION
_HEDGEHOGMESSAGE.fields_by_name['process_exit_update'].message_type = hedgehog_dot_protocol_dot_proto_dot_process__pb2._PROCESSEXITUPDATE
_HEDGEHOGMESSAGE.oneofs_by_name['payload'].fields.append(
  _HEDGEHOGMESSAGE.fields_by_name['acknowledgement'])
_HEDGEHOGMESSAGE.fields_by_name['acknowledgement'].containing_oneof = _HEDGEHOGMESSAGE.oneofs_by_name['payload']
_HEDGEHOGMESSAGE.oneofs_by_name['payload'].fields.append(
  _HEDGEHOGMESSAGE.fields_by_name['io_action'])
_HEDGEHOGMESSAGE.fields_by_name['io_action'].containing_oneof = _HEDGEHOGMESSAGE.oneofs_by_name['payload']
_HEDGEHOGMESSAGE.oneofs_by_name['payload'].fields.append(
  _HEDGEHOGMESSAGE.fields_by_name['io_command_message'])
_HEDGEHOGMESSAGE.fields_by_name['io_command_message'].containing_oneof = _HEDGEHOGMESSAGE.oneofs_by_name['payload']
_HEDGEHOGMESSAGE.oneofs_by_name['payload'].fields.append(
  _HEDGEHOGMESSAGE.fields_by_name['analog_message'])
_HEDGEHOGMESSAGE.fields_by_name['analog_message'].containing_oneof = _HEDGEHOGMESSAGE.oneofs_by_name['payload']
_HEDGEHOGMESSAGE.oneofs_by_name['payload'].fields.append(
  _HEDGEHOGMESSAGE.fields_by_name['digital_message'])
_HEDGEHOGMESSAGE.fields_by_name['digital_message'].containing_oneof = _HEDGEHOGMESSAGE.oneofs_by_name['payload']
_HEDGEHOGMESSAGE.oneofs_by_name['payload'].fields.append(
  _HEDGEHOGMESSAGE.fields_by_name['motor_action'])
_HEDGEHOGMESSAGE.fields_by_name['motor_action'].containing_oneof = _HEDGEHOGMESSAGE.oneofs_by_name['payload']
_HEDGEHOGMESSAGE.oneofs_by_name['payload'].fields.append(
  _HEDGEHOGMESSAGE.fields_by_name['motor_command_message'])
_HEDGEHOGMESSAGE.fields_by_name['motor_command_message'].containing_oneof = _HEDGEHOGMESSAGE.oneofs_by_name['payload']
_HEDGEHOGMESSAGE.oneofs_by_name['payload'].fields.append(
  _HEDGEHOGMESSAGE.fields_by_name['motor_state_message'])
_HEDGEHOGMESSAGE.fields_by_name['motor_state_message'].containing_oneof = _HEDGEHOGMESSAGE.oneofs_by_name['payload']
_HEDGEHOGMESSAGE.oneofs_by_name['payload'].fields.append(
  _HEDGEHOGMESSAGE.fields_by_name['motor_set_position_action'])
_HEDGEHOGMESSAGE.fields_by_name['motor_set_position_action'].containing_oneof = _HEDGEHOGMESSAGE.oneofs_by_name['payload']
_HEDGEHOGMESSAGE.oneofs_by_name['payload'].fields.append(
  _HEDGEHOGMESSAGE.fields_by_name['servo_action'])
_HEDGEHOGMESSAGE.fields_by_name['servo_action'].containing_oneof = _HEDGEHOGMESSAGE.oneofs_by_name['payload']
_HEDGEHOGMESSAGE.oneofs_by_name['payload'].fields.append(
  _HEDGEHOGMESSAGE.fields_by_name['servo_command_message'])
_HEDGEHOGMESSAGE.fields_by_name['servo_command_message'].containing_oneof = _HEDGEHOGMESSAGE.oneofs_by_name['payload']
_HEDGEHOGMESSAGE.oneofs_by_name['payload'].fields.append(
  _HEDGEHOGMESSAGE.fields_by_name['process_execute_action'])
_HEDGEHOGMESSAGE.fields_by_name['process_execute_action'].containing_oneof = _HEDGEHOGMESSAGE.oneofs_by_name['payload']
_HEDGEHOGMESSAGE.oneofs_by_name['payload'].fields.append(
  _HEDGEHOGMESSAGE.fields_by_name['process_execute_reply'])
_HEDGEHOGMESSAGE.fields_by_name['process_execute_reply'].containing_oneof = _HEDGEHOGMESSAGE.oneofs_by_name['payload']
_HEDGEHOGMESSAGE.oneofs_by_name['payload'].fields.append(
  _HEDGEHOGMESSAGE.fields_by_name['process_stream_message'])
_HEDGEHOGMESSAGE.fields_by_name['process_stream_message'].containing_oneof = _HEDGEHOGMESSAGE.oneofs_by_name['payload']
_HEDGEHOGMESSAGE.oneofs_by_name['payload'].fields.append(
  _HEDGEHOGMESSAGE.fields_by_name['process_signal_action'])
_HEDGEHOGMESSAGE.fields_by_name['process_signal_action'].containing_oneof = _HEDGEHOGMESSAGE.oneofs_by_name['payload']
_HEDGEHOGMESSAGE.oneofs_by_name['payload'].fields.append(
  _HEDGEHOGMESSAGE.fields_by_name['process_exit_update'])
_HEDGEHOGMESSAGE.fields_by_name['process_exit_update'].containing_oneof = _HEDGEHOGMESSAGE.oneofs_by_name['payload']
DESCRIPTOR.message_types_by_name['HedgehogMessage'] = _HEDGEHOGMESSAGE

HedgehogMessage = _reflection.GeneratedProtocolMessageType('HedgehogMessage', (_message.Message,), dict(
  DESCRIPTOR = _HEDGEHOGMESSAGE,
  __module__ = 'hedgehog.protocol.proto.hedgehog_pb2'
  # @@protoc_insertion_point(class_scope:hedgehog.protocol.proto.HedgehogMessage)
  ))
_sym_db.RegisterMessage(HedgehogMessage)


DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(descriptor_pb2.FileOptions(), _b('\n\037at.pria.hedgehog.protocol.protoB\tHedgehogP'))
# @@protoc_insertion_point(module_scope)



from typing import Sequence, TypeVar

from hedgehog.protocol import CommSide
from hedgehog.utils.zmq.socket import Socket
from . import RawMessage, Message, RawPayload, Payload, \
    Header, DelimitedMsg, RawMsgs, Msgs, RawMsg, Msg


T = TypeVar('T')


def _rindex(mylist: Sequence[T], x: T) -> int:
    """Index of the last occurrence of x in the sequence."""
    return len(mylist) - mylist[::-1].index(x) - 1


def raw_to_delimited(header: Header, raw_payload: RawPayload) -> DelimitedMsg:
    """
    Returns a message consisting of header frames, delimiter frame, and payload frames.
    The payload frames may be given as sequences of bytes (raw) or as `Message`s.
    """
    return tuple(header) + (b'',) + tuple(raw_payload)


def to_delimited(header: Header, payload: Payload, side: CommSide) -> DelimitedMsg:
    """
    Returns a message consisting of header frames, delimiter frame, and payload frames.
    The payload frames may be given as sequences of bytes (raw) or as `Message`s.
    """
    return raw_to_delimited(header, [side.serialize(msg) for msg in payload])


def raw_from_delimited(msgs: DelimitedMsg) -> RawMsgs:
    """
    From a message consisting of header frames, delimiter frame, and payload frames, return a tuple `(header, payload)`.
    The payload frames may be returned as sequences of bytes (raw) or as `Message`s.
    """
    delim = _rindex(msgs, b'')
    return tuple(msgs[:delim]), tuple(msgs[delim + 1:])


def from_delimited(msgs: DelimitedMsg, side: CommSide) -> Msgs:
    """
    From a message consisting of header frames, delimiter frame, and payload frames, return a tuple `(header, payload)`.
    The payload frames may be returned as sequences of bytes (raw) or as `Message`s.
    """
    header, raw_payload = raw_from_delimited(msgs)
    return header, tuple(side.parse(msg_raw) for msg_raw in raw_payload)


class DealerRouterMixin(object):
    """
    A mixin for ZMQ dealer & router sockets used to send delimited & Hedgehog-encoded messages.

    This mixin defines methods to send/receive single/multipart binary/hedgehog messages on dealer & router sockets.
    For example, `send_msg` send a single hedgehog message, while `recv_msgs_raw` receives a multipart binary message.
    All these methods use a header (one or more binary frames) followed by a delimiter (one empty frame). `send` methods
    accept a header parameter before the payload, `recv` methods return the header and payload as a tuple.
    """
    def send_msg(self, header: Header, msg: Message):
        self.send_msgs(header, [msg])

    def recv_msg(self) -> Msg:
        header, [msg] = self.recv_msgs()
        return header, msg

    def send_msgs(self, header: Header, msgs: Payload):
        self.send_msgs_raw(header, [self.side.serialize(msg) for msg in msgs])

    def recv_msgs(self) -> Msgs:
        header, msgs_raw = self.recv_msgs_raw()
        return header, tuple(self.side.parse(msg_raw) for msg_raw in msgs_raw)

    def send_msg_raw(self, header: Header, msg_raw: RawMessage):
        self.send_msgs_raw(header, [msg_raw])

    def recv_msg_raw(self) -> RawMsg:
        header, [msg_raw] = self.recv_msgs_raw()
        return header, msg_raw

    def send_msgs_raw(self, header: Header, msgs_raw: RawPayload):
        self.send_multipart(raw_to_delimited(header, msgs_raw))

    def recv_msgs_raw(self) -> RawMsgs:
        return raw_from_delimited(self.recv_multipart())


class DealerRouterSocket(DealerRouterMixin, Socket):
    side = None

    def __init__(self, *args, side: CommSide, **kwargs) -> None:
        super(DealerRouterSocket, self).__init__(*args, **kwargs)
        self.side = side


class ReqMixin(object):
    """
    A mixin for ZMQ req sockets used to send Hedgehog-encoded messages.

    This mixin defines methods to send/receive single/multipart binary/hedgehog messages on req sockets.
    For example, `send_msg` send a single hedgehog message, while `recv_msgs_raw` receives a multipart binary message.
    All these methods use a delimiter (one empty frame), implicitly added by the req socket.
    """

    def send_msg(self, msg: Message):
        self.send_msgs([msg])

    def recv_msg(self) -> Message:
        [msg] = self.recv_msgs()
        return msg

    def send_msgs(self, msgs: Payload):
        self.send_msgs_raw([self.side.serialize(msg) for msg in msgs])

    def recv_msgs(self) -> Payload:
        return tuple(self.side.parse(msg_raw) for msg_raw in self.recv_msgs_raw())

    def send_msg_raw(self, msg_raw: RawMessage):
        self.send_msgs_raw([msg_raw])

    def recv_msg_raw(self) -> RawMessage:
        [msg_raw] = self.recv_msgs_raw()
        return msg_raw

    def send_msgs_raw(self, msgs_raw: RawPayload):
        self.send_multipart(msgs_raw)

    def recv_msgs_raw(self) -> RawPayload:
        return self.recv_multipart()


class ReqSocket(ReqMixin, Socket):
    side = None

    def __init__(self, *args, side: CommSide, **kwargs) -> None:
        super(ReqSocket, self).__init__(*args, **kwargs)
        self.side = side



from . import ReplyMsg, SimpleMessage
from hedgehog.protocol.proto import ack_pb2
from hedgehog.protocol.proto.ack_pb2 import OK, UNKNOWN_COMMAND, INVALID_COMMAND, UNSUPPORTED_COMMAND, FAILED_COMMAND


@ReplyMsg.message(ack_pb2.Acknowledgement, 'acknowledgement')
class Acknowledgement(SimpleMessage):
    def __init__(self, code: int=OK, message: str='') -> None:
        self.code = code
        self.message = message

    @classmethod
    def _parse(cls, msg: ack_pb2.Acknowledgement) -> 'Acknowledgement':
        code = msg.code
        message = msg.message
        return cls(code, message)

    def _serialize(self, msg: ack_pb2.Acknowledgement) -> None:
        msg.code = self.code
        msg.message = self.message



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: hedgehog/protocol/proto/process.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='hedgehog/protocol/proto/process.proto',
  package='hedgehog.protocol.proto',
  syntax='proto3',
  serialized_pb=_b('\n%hedgehog/protocol/proto/process.proto\x12\x17hedgehog.protocol.proto\"9\n\x14ProcessExecuteAction\x12\x0c\n\x04\x61rgs\x18\x02 \x03(\t\x12\x13\n\x0bworking_dir\x18\x01 \x01(\t\"\"\n\x13ProcessExecuteReply\x12\x0b\n\x03pid\x18\x01 \x01(\r\"j\n\x14ProcessStreamMessage\x12\x0b\n\x03pid\x18\x01 \x01(\r\x12\x36\n\x06\x66ileno\x18\x02 \x01(\x0e\x32&.hedgehog.protocol.proto.ProcessFileno\x12\r\n\x05\x63hunk\x18\x03 \x01(\x0c\"2\n\x13ProcessSignalAction\x12\x0b\n\x03pid\x18\x01 \x01(\r\x12\x0e\n\x06signal\x18\x02 \x01(\r\"3\n\x11ProcessExitUpdate\x12\x0b\n\x03pid\x18\x01 \x01(\r\x12\x11\n\texit_code\x18\x02 \x01(\x05*2\n\rProcessFileno\x12\t\n\x05STDIN\x10\x00\x12\n\n\x06STDOUT\x10\x01\x12\n\n\x06STDERR\x10\x02\x42+\n\x1f\x61t.pria.hedgehog.protocol.protoB\x08ProcessPb\x06proto3')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

_PROCESSFILENO = _descriptor.EnumDescriptor(
  name='ProcessFileno',
  full_name='hedgehog.protocol.proto.ProcessFileno',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='STDIN', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='STDOUT', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='STDERR', index=2, number=2,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=374,
  serialized_end=424,
)
_sym_db.RegisterEnumDescriptor(_PROCESSFILENO)

ProcessFileno = enum_type_wrapper.EnumTypeWrapper(_PROCESSFILENO)
STDIN = 0
STDOUT = 1
STDERR = 2



_PROCESSEXECUTEACTION = _descriptor.Descriptor(
  name='ProcessExecuteAction',
  full_name='hedgehog.protocol.proto.ProcessExecuteAction',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='args', full_name='hedgehog.protocol.proto.ProcessExecuteAction.args', index=0,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='working_dir', full_name='hedgehog.protocol.proto.ProcessExecuteAction.working_dir', index=1,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=66,
  serialized_end=123,
)


_PROCESSEXECUTEREPLY = _descriptor.Descriptor(
  name='ProcessExecuteReply',
  full_name='hedgehog.protocol.proto.ProcessExecuteReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='pid', full_name='hedgehog.protocol.proto.ProcessExecuteReply.pid', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=125,
  serialized_end=159,
)


_PROCESSSTREAMMESSAGE = _descriptor.Descriptor(
  name='ProcessStreamMessage',
  full_name='hedgehog.protocol.proto.ProcessStreamMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='pid', full_name='hedgehog.protocol.proto.ProcessStreamMessage.pid', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='fileno', full_name='hedgehog.protocol.proto.ProcessStreamMessage.fileno', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='chunk', full_name='hedgehog.protocol.proto.ProcessStreamMessage.chunk', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=161,
  serialized_end=267,
)


_PROCESSSIGNALACTION = _descriptor.Descriptor(
  name='ProcessSignalAction',
  full_name='hedgehog.protocol.proto.ProcessSignalAction',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='pid', full_name='hedgehog.protocol.proto.ProcessSignalAction.pid', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='signal', full_name='hedgehog.protocol.proto.ProcessSignalAction.signal', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=269,
  serialized_end=319,
)


_PROCESSEXITUPDATE = _descriptor.Descriptor(
  name='ProcessExitUpdate',
  full_name='hedgehog.protocol.proto.ProcessExitUpdate',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='pid', full_name='hedgehog.protocol.proto.ProcessExitUpdate.pid', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='exit_code', full_name='hedgehog.protocol.proto.ProcessExitUpdate.exit_code', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=321,
  serialized_end=372,
)

_PROCESSSTREAMMESSAGE.fields_by_name['fileno'].enum_type = _PROCESSFILENO
DESCRIPTOR.message_types_by_name['ProcessExecuteAction'] = _PROCESSEXECUTEACTION
DESCRIPTOR.message_types_by_name['ProcessExecuteReply'] = _PROCESSEXECUTEREPLY
DESCRIPTOR.message_types_by_name['ProcessStreamMessage'] = _PROCESSSTREAMMESSAGE
DESCRIPTOR.message_types_by_name['ProcessSignalAction'] = _PROCESSSIGNALACTION
DESCRIPTOR.message_types_by_name['ProcessExitUpdate'] = _PROCESSEXITUPDATE
DESCRIPTOR.enum_types_by_name['ProcessFileno'] = _PROCESSFILENO

ProcessExecuteAction = _reflection.GeneratedProtocolMessageType('ProcessExecuteAction', (_message.Message,), dict(
  DESCRIPTOR = _PROCESSEXECUTEACTION,
  __module__ = 'hedgehog.protocol.proto.process_pb2'
  # @@protoc_insertion_point(class_scope:hedgehog.protocol.proto.ProcessExecuteAction)
  ))
_sym_db.RegisterMessage(ProcessExecuteAction)

ProcessExecuteReply = _reflection.GeneratedProtocolMessageType('ProcessExecuteReply', (_message.Message,), dict(
  DESCRIPTOR = _PROCESSEXECUTEREPLY,
  __module__ = 'hedgehog.protocol.proto.process_pb2'
  # @@protoc_insertion_point(class_scope:hedgehog.protocol.proto.ProcessExecuteReply)
  ))
_sym_db.RegisterMessage(ProcessExecuteReply)

ProcessStreamMessage = _reflection.GeneratedProtocolMessageType('ProcessStreamMessage', (_message.Message,), dict(
  DESCRIPTOR = _PROCESSSTREAMMESSAGE,
  __module__ = 'hedgehog.protocol.proto.process_pb2'
  # @@protoc_insertion_point(class_scope:hedgehog.protocol.proto.ProcessStreamMessage)
  ))
_sym_db.RegisterMessage(ProcessStreamMessage)

ProcessSignalAction = _reflection.GeneratedProtocolMessageType('ProcessSignalAction', (_message.Message,), dict(
  DESCRIPTOR = _PROCESSSIGNALACTION,
  __module__ = 'hedgehog.protocol.proto.process_pb2'
  # @@protoc_insertion_point(class_scope:hedgehog.protocol.proto.ProcessSignalAction)
  ))
_sym_db.RegisterMessage(ProcessSignalAction)

ProcessExitUpdate = _reflection.GeneratedProtocolMessageType('ProcessExitUpdate', (_message.Message,), dict(
  DESCRIPTOR = _PROCESSEXITUPDATE,
  __module__ = 'hedgehog.protocol.proto.process_pb2'
  # @@protoc_insertion_point(class_scope:hedgehog.protocol.proto.ProcessExitUpdate)
  ))
_sym_db.RegisterMessage(ProcessExitUpdate)


DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(descriptor_pb2.FileOptions(), _b('\n\037at.pria.hedgehog.protocol.protoB\010ProcessP'))
# @@protoc_insertion_point(module_scope)



from typing import Sequence, Tuple

from hedgehog.utils import protobuf

from .messages import ContainerMessage, Message, RequestMsg, ReplyMsg
from .errors import UnknownCommandError


# a single protobuf-encoded message
RawMessage = bytes
# a single decoded message
#Message
# a sequence of raw messages
RawPayload = Sequence[RawMessage]
# a sequence of decoded messages
Payload = Sequence[Message]

# sequence of zeromq header frames
Header = Sequence[bytes]
# a multipart ZeroMQ message with headers, delimiter and payload
DelimitedMsg = Sequence[bytes]
# a decoded message with headers and raw payload
RawMsgs = Tuple[Header, RawPayload]
# a decoded message with headers and payload
Msgs = Tuple[Header, Payload]
# a decoded message with headers and a single raw message
RawMsg = Tuple[Header, RawMessage]
# a decoded message with headers and a single message
Msg = Tuple[Header, Message]


class CommSide(object):
    def __init__(self, receiver: ContainerMessage, sender: protobuf.ContainerMessage) -> None:
        self.receiver = receiver
        self.sender = sender

    def parse(self, data: RawMessage) -> Message:
        """Parses a binary protobuf message into a Message object."""
        try:
            return self.receiver.parse(data)
        except KeyError as err:
            raise UnknownCommandError() from err

    def serialize(self, msg: Message) -> RawMessage:
        """Serializes a Message object into a binary protobuf message"""
        return self.sender.serialize(msg)


ServerSide = CommSide(RequestMsg, ReplyMsg)
ClientSide = CommSide(ReplyMsg, RequestMsg)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: hedgehog/protocol/proto/digital.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='hedgehog/protocol/proto/digital.proto',
  package='hedgehog.protocol.proto',
  syntax='proto3',
  serialized_pb=_b('\n%hedgehog/protocol/proto/digital.proto\x12\x17hedgehog.protocol.proto\"\x1e\n\x0e\x44igitalRequest\x12\x0c\n\x04port\x18\x01 \x01(\r\",\n\rDigitalUpdate\x12\x0c\n\x04port\x18\x01 \x01(\r\x12\r\n\x05value\x18\x02 \x01(\x08\"B\n\x12\x44igitalStateAction\x12\x0c\n\x04port\x18\x01 \x01(\r\x12\x0e\n\x06pullup\x18\x02 \x01(\x08\x12\x0e\n\x06output\x18\x03 \x01(\x08\",\n\rDigitalAction\x12\x0c\n\x04port\x18\x01 \x01(\r\x12\r\n\x05level\x18\x02 \x01(\x08\x62\x06proto3')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_DIGITALREQUEST = _descriptor.Descriptor(
  name='DigitalRequest',
  full_name='hedgehog.protocol.proto.DigitalRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='port', full_name='hedgehog.protocol.proto.DigitalRequest.port', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=66,
  serialized_end=96,
)


_DIGITALUPDATE = _descriptor.Descriptor(
  name='DigitalUpdate',
  full_name='hedgehog.protocol.proto.DigitalUpdate',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='port', full_name='hedgehog.protocol.proto.DigitalUpdate.port', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='value', full_name='hedgehog.protocol.proto.DigitalUpdate.value', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=98,
  serialized_end=142,
)


_DIGITALSTATEACTION = _descriptor.Descriptor(
  name='DigitalStateAction',
  full_name='hedgehog.protocol.proto.DigitalStateAction',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='port', full_name='hedgehog.protocol.proto.DigitalStateAction.port', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='pullup', full_name='hedgehog.protocol.proto.DigitalStateAction.pullup', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='output', full_name='hedgehog.protocol.proto.DigitalStateAction.output', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=144,
  serialized_end=210,
)


_DIGITALACTION = _descriptor.Descriptor(
  name='DigitalAction',
  full_name='hedgehog.protocol.proto.DigitalAction',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='port', full_name='hedgehog.protocol.proto.DigitalAction.port', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='level', full_name='hedgehog.protocol.proto.DigitalAction.level', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=212,
  serialized_end=256,
)

DESCRIPTOR.message_types_by_name['DigitalRequest'] = _DIGITALREQUEST
DESCRIPTOR.message_types_by_name['DigitalUpdate'] = _DIGITALUPDATE
DESCRIPTOR.message_types_by_name['DigitalStateAction'] = _DIGITALSTATEACTION
DESCRIPTOR.message_types_by_name['DigitalAction'] = _DIGITALACTION

DigitalRequest = _reflection.GeneratedProtocolMessageType('DigitalRequest', (_message.Message,), dict(
  DESCRIPTOR = _DIGITALREQUEST,
  __module__ = 'hedgehog.protocol.proto.digital_pb2'
  # @@protoc_insertion_point(class_scope:hedgehog.protocol.proto.DigitalRequest)
  ))
_sym_db.RegisterMessage(DigitalRequest)

DigitalUpdate = _reflection.GeneratedProtocolMessageType('DigitalUpdate', (_message.Message,), dict(
  DESCRIPTOR = _DIGITALUPDATE,
  __module__ = 'hedgehog.protocol.proto.digital_pb2'
  # @@protoc_insertion_point(class_scope:hedgehog.protocol.proto.DigitalUpdate)
  ))
_sym_db.RegisterMessage(DigitalUpdate)

DigitalStateAction = _reflection.GeneratedProtocolMessageType('DigitalStateAction', (_message.Message,), dict(
  DESCRIPTOR = _DIGITALSTATEACTION,
  __module__ = 'hedgehog.protocol.proto.digital_pb2'
  # @@protoc_insertion_point(class_scope:hedgehog.protocol.proto.DigitalStateAction)
  ))
_sym_db.RegisterMessage(DigitalStateAction)

DigitalAction = _reflection.GeneratedProtocolMessageType('DigitalAction', (_message.Message,), dict(
  DESCRIPTOR = _DIGITALACTION,
  __module__ = 'hedgehog.protocol.proto.digital_pb2'
  # @@protoc_insertion_point(class_scope:hedgehog.protocol.proto.DigitalAction)
  ))
_sym_db.RegisterMessage(DigitalAction)


# @@protoc_insertion_point(module_scope)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: hedgehog/protocol/proto/ack.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='hedgehog/protocol/proto/ack.proto',
  package='hedgehog.protocol.proto',
  syntax='proto3',
  serialized_pb=_b('\n!hedgehog/protocol/proto/ack.proto\x12\x17hedgehog.protocol.proto\"^\n\x0f\x41\x63knowledgement\x12:\n\x04\x63ode\x18\x01 \x01(\x0e\x32,.hedgehog.protocol.proto.AcknowledgementCode\x12\x0f\n\x07message\x18\x02 \x01(\t*t\n\x13\x41\x63knowledgementCode\x12\x06\n\x02OK\x10\x00\x12\x13\n\x0fUNKNOWN_COMMAND\x10\x01\x12\x13\n\x0fINVALID_COMMAND\x10\x02\x12\x17\n\x13UNSUPPORTED_COMMAND\x10\x03\x12\x12\n\x0e\x46\x41ILED_COMMAND\x10\x04\x42\'\n\x1f\x61t.pria.hedgehog.protocol.protoB\x04\x41\x63kPb\x06proto3')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

_ACKNOWLEDGEMENTCODE = _descriptor.EnumDescriptor(
  name='AcknowledgementCode',
  full_name='hedgehog.protocol.proto.AcknowledgementCode',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='OK', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN_COMMAND', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='INVALID_COMMAND', index=2, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='UNSUPPORTED_COMMAND', index=3, number=3,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FAILED_COMMAND', index=4, number=4,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=158,
  serialized_end=274,
)
_sym_db.RegisterEnumDescriptor(_ACKNOWLEDGEMENTCODE)

AcknowledgementCode = enum_type_wrapper.EnumTypeWrapper(_ACKNOWLEDGEMENTCODE)
OK = 0
UNKNOWN_COMMAND = 1
INVALID_COMMAND = 2
UNSUPPORTED_COMMAND = 3
FAILED_COMMAND = 4



_ACKNOWLEDGEMENT = _descriptor.Descriptor(
  name='Acknowledgement',
  full_name='hedgehog.protocol.proto.Acknowledgement',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='code', full_name='hedgehog.protocol.proto.Acknowledgement.code', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='message', full_name='hedgehog.protocol.proto.Acknowledgement.message', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=62,
  serialized_end=156,
)

_ACKNOWLEDGEMENT.fields_by_name['code'].enum_type = _ACKNOWLEDGEMENTCODE
DESCRIPTOR.message_types_by_name['Acknowledgement'] = _ACKNOWLEDGEMENT
DESCRIPTOR.enum_types_by_name['AcknowledgementCode'] = _ACKNOWLEDGEMENTCODE

Acknowledgement = _reflection.GeneratedProtocolMessageType('Acknowledgement', (_message.Message,), dict(
  DESCRIPTOR = _ACKNOWLEDGEMENT,
  __module__ = 'hedgehog.protocol.proto.ack_pb2'
  # @@protoc_insertion_point(class_scope:hedgehog.protocol.proto.Acknowledgement)
  ))
_sym_db.RegisterMessage(Acknowledgement)


DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(descriptor_pb2.FileOptions(), _b('\n\037at.pria.hedgehog.protocol.protoB\004AckP'))
# @@protoc_insertion_point(module_scope)



from typing import Union

from . import RequestMsg, ReplyMsg, Message, SimpleMessage
from hedgehog.protocol.errors import InvalidCommandError
from hedgehog.protocol.proto import servo_pb2
from hedgehog.protocol.proto.subscription_pb2 import Subscription
from hedgehog.utils import protobuf


@RequestMsg.message(servo_pb2.ServoAction, 'servo_action')
class Action(SimpleMessage):
    def __init__(self, port: int, active: bool, position: int=None) -> None:
        if active and position is None:
            raise InvalidCommandError("position must be given when activating servo")
        self.port = port
        self.active = active
        self.position = position if active else None

    @classmethod
    def _parse(cls, msg: servo_pb2.ServoAction) -> 'Action':
        port = msg.port
        active = msg.active
        position = msg.position
        return cls(port, active, position)

    def _serialize(self, msg: servo_pb2.ServoAction) -> None:
        msg.port = self.port
        msg.active = self.active
        if self.position is not None:
            msg.position = self.position


@protobuf.message(servo_pb2.ServoCommandMessage, 'servo_command_message', fields=('port',))
class CommandRequest(Message):
    def __init__(self, port: int) -> None:
        self.port = port

    def _serialize(self, msg: servo_pb2.ServoCommandMessage) -> None:
        msg.port = self.port


@protobuf.message(servo_pb2.ServoCommandMessage, 'servo_command_message', fields=('port', 'subscription'))
class CommandSubscribe(Message):
    def __init__(self, port: int, subscription: Subscription) -> None:
        self.port = port
        self.subscription = subscription

    def _serialize(self, msg: servo_pb2.ServoCommandMessage) -> None:
        msg.port = self.port
        msg.subscription.CopyFrom(self.subscription)


@RequestMsg.parser('servo_command_message')
def _parse_command_request(msg: servo_pb2.ServoCommandMessage) -> Union[CommandRequest, CommandSubscribe]:
    port = msg.port
    subscription = msg.subscription if msg.HasField('subscription') else None
    if subscription is None:
        return CommandRequest(port)
    else:
        return CommandSubscribe(port, subscription)


@protobuf.message(servo_pb2.ServoCommandMessage, 'servo_command_message', fields=('port', 'active', 'position'))
class CommandReply(Message):
    def __init__(self, port: int, active: bool, position: int) -> None:
        self.port = port
        self.active = active
        self.position = position if active else None

    def _serialize(self, msg: servo_pb2.ServoCommandMessage) -> None:
        msg.port = self.port
        msg.active = self.active
        if self.position is not None:
            msg.position = self.position


@protobuf.message(servo_pb2.ServoCommandMessage, 'servo_command_message')
class CommandUpdate(Message):
    is_async = True

    def __init__(self, port: int, active: bool, position: int, subscription: Subscription) -> None:
        self.port = port
        self.active = active
        self.position = position if active else None
        self.subscription = subscription

    def _serialize(self, msg: servo_pb2.ServoCommandMessage) -> None:
        msg.port = self.port
        msg.active = self.active
        if self.position is not None:
            msg.position = self.position
        msg.subscription.CopyFrom(self.subscription)


@ReplyMsg.parser('servo_command_message')
def _parse_command_reply(msg: servo_pb2.ServoCommandMessage) -> Union[CommandReply, CommandUpdate]:
    port = msg.port
    active = msg.active
    position = msg.position
    subscription = msg.subscription if msg.HasField('subscription') else None
    if subscription is None:
        return CommandReply(port, active, position)
    else:
        return CommandUpdate(port, active, position, subscription)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: hedgehog/protocol/proto/io.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from hedgehog.protocol.proto import subscription_pb2 as hedgehog_dot_protocol_dot_proto_dot_subscription__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='hedgehog/protocol/proto/io.proto',
  package='hedgehog.protocol.proto',
  syntax='proto3',
  serialized_pb=_b('\n hedgehog/protocol/proto/io.proto\x12\x17hedgehog.protocol.proto\x1a*hedgehog/protocol/proto/subscription.proto\"I\n\x08IOAction\x12\x0c\n\x04port\x18\x01 \x01(\r\x12/\n\x05\x66lags\x18\x02 \x01(\x0e\x32 .hedgehog.protocol.proto.IOFlags\"\x8e\x01\n\x10IOCommandMessage\x12\x0c\n\x04port\x18\x01 \x01(\r\x12/\n\x05\x66lags\x18\x02 \x01(\x0e\x32 .hedgehog.protocol.proto.IOFlags\x12;\n\x0csubscription\x18\x03 \x01(\x0b\x32%.hedgehog.protocol.proto.Subscription\"i\n\rAnalogMessage\x12\x0c\n\x04port\x18\x01 \x01(\r\x12\r\n\x05value\x18\x02 \x01(\r\x12;\n\x0csubscription\x18\x03 \x01(\x0b\x32%.hedgehog.protocol.proto.Subscription\"j\n\x0e\x44igitalMessage\x12\x0c\n\x04port\x18\x01 \x01(\r\x12\r\n\x05value\x18\x02 \x01(\x08\x12;\n\x0csubscription\x18\x03 \x01(\x0b\x32%.hedgehog.protocol.proto.Subscription*\x97\x01\n\x07IOFlags\x12\x12\n\x0eINPUT_FLOATING\x10\x00\x12\x10\n\x0cINPUT_PULLUP\x10\x02\x12\x12\n\x0eINPUT_PULLDOWN\x10\x04\x12\x0e\n\nOUTPUT_OFF\x10\x01\x12\r\n\tOUTPUT_ON\x10\t\x12\n\n\x06OUTPUT\x10\x01\x12\n\n\x06PULLUP\x10\x02\x12\x0c\n\x08PULLDOWN\x10\x04\x12\t\n\x05LEVEL\x10\x08\x1a\x02\x10\x01\x42&\n\x1f\x61t.pria.hedgehog.protocol.protoB\x03IOPb\x06proto3')
  ,
  dependencies=[hedgehog_dot_protocol_dot_proto_dot_subscription__pb2.DESCRIPTOR,])
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

_IOFLAGS = _descriptor.EnumDescriptor(
  name='IOFlags',
  full_name='hedgehog.protocol.proto.IOFlags',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='INPUT_FLOATING', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='INPUT_PULLUP', index=1, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='INPUT_PULLDOWN', index=2, number=4,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='OUTPUT_OFF', index=3, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='OUTPUT_ON', index=4, number=9,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='OUTPUT', index=5, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PULLUP', index=6, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PULLDOWN', index=7, number=4,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LEVEL', index=8, number=8,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=_descriptor._ParseOptions(descriptor_pb2.EnumOptions(), _b('\020\001')),
  serialized_start=541,
  serialized_end=692,
)
_sym_db.RegisterEnumDescriptor(_IOFLAGS)

IOFlags = enum_type_wrapper.EnumTypeWrapper(_IOFLAGS)
INPUT_FLOATING = 0
INPUT_PULLUP = 2
INPUT_PULLDOWN = 4
OUTPUT_OFF = 1
OUTPUT_ON = 9
OUTPUT = 1
PULLUP = 2
PULLDOWN = 4
LEVEL = 8



_IOACTION = _descriptor.Descriptor(
  name='IOAction',
  full_name='hedgehog.protocol.proto.IOAction',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='port', full_name='hedgehog.protocol.proto.IOAction.port', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='flags', full_name='hedgehog.protocol.proto.IOAction.flags', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=105,
  serialized_end=178,
)


_IOCOMMANDMESSAGE = _descriptor.Descriptor(
  name='IOCommandMessage',
  full_name='hedgehog.protocol.proto.IOCommandMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='port', full_name='hedgehog.protocol.proto.IOCommandMessage.port', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='flags', full_name='hedgehog.protocol.proto.IOCommandMessage.flags', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='subscription', full_name='hedgehog.protocol.proto.IOCommandMessage.subscription', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=181,
  serialized_end=323,
)


_ANALOGMESSAGE = _descriptor.Descriptor(
  name='AnalogMessage',
  full_name='hedgehog.protocol.proto.AnalogMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='port', full_name='hedgehog.protocol.proto.AnalogMessage.port', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='value', full_name='hedgehog.protocol.proto.AnalogMessage.value', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='subscription', full_name='hedgehog.protocol.proto.AnalogMessage.subscription', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=325,
  serialized_end=430,
)


_DIGITALMESSAGE = _descriptor.Descriptor(
  name='DigitalMessage',
  full_name='hedgehog.protocol.proto.DigitalMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='port', full_name='hedgehog.protocol.proto.DigitalMessage.port', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='value', full_name='hedgehog.protocol.proto.DigitalMessage.value', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='subscription', full_name='hedgehog.protocol.proto.DigitalMessage.subscription', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=432,
  serialized_end=538,
)

_IOACTION.fields_by_name['flags'].enum_type = _IOFLAGS
_IOCOMMANDMESSAGE.fields_by_name['flags'].enum_type = _IOFLAGS
_IOCOMMANDMESSAGE.fields_by_name['subscription'].message_type = hedgehog_dot_protocol_dot_proto_dot_subscription__pb2._SUBSCRIPTION
_ANALOGMESSAGE.fields_by_name['subscription'].message_type = hedgehog_dot_protocol_dot_proto_dot_subscription__pb2._SUBSCRIPTION
_DIGITALMESSAGE.fields_by_name['subscription'].message_type = hedgehog_dot_protocol_dot_proto_dot_subscription__pb2._SUBSCRIPTION
DESCRIPTOR.message_types_by_name['IOAction'] = _IOACTION
DESCRIPTOR.message_types_by_name['IOCommandMessage'] = _IOCOMMANDMESSAGE
DESCRIPTOR.message_types_by_name['AnalogMessage'] = _ANALOGMESSAGE
DESCRIPTOR.message_types_by_name['DigitalMessage'] = _DIGITALMESSAGE
DESCRIPTOR.enum_types_by_name['IOFlags'] = _IOFLAGS

IOAction = _reflection.GeneratedProtocolMessageType('IOAction', (_message.Message,), dict(
  DESCRIPTOR = _IOACTION,
  __module__ = 'hedgehog.protocol.proto.io_pb2'
  # @@protoc_insertion_point(class_scope:hedgehog.protocol.proto.IOAction)
  ))
_sym_db.RegisterMessage(IOAction)

IOCommandMessage = _reflection.GeneratedProtocolMessageType('IOCommandMessage', (_message.Message,), dict(
  DESCRIPTOR = _IOCOMMANDMESSAGE,
  __module__ = 'hedgehog.protocol.proto.io_pb2'
  # @@protoc_insertion_point(class_scope:hedgehog.protocol.proto.IOCommandMessage)
  ))
_sym_db.RegisterMessage(IOCommandMessage)

AnalogMessage = _reflection.GeneratedProtocolMessageType('AnalogMessage', (_message.Message,), dict(
  DESCRIPTOR = _ANALOGMESSAGE,
  __module__ = 'hedgehog.protocol.proto.io_pb2'
  # @@protoc_insertion_point(class_scope:hedgehog.protocol.proto.AnalogMessage)
  ))
_sym_db.RegisterMessage(AnalogMessage)

DigitalMessage = _reflection.GeneratedProtocolMessageType('DigitalMessage', (_message.Message,), dict(
  DESCRIPTOR = _DIGITALMESSAGE,
  __module__ = 'hedgehog.protocol.proto.io_pb2'
  # @@protoc_insertion_point(class_scope:hedgehog.protocol.proto.DigitalMessage)
  ))
_sym_db.RegisterMessage(DigitalMessage)


DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(descriptor_pb2.FileOptions(), _b('\n\037at.pria.hedgehog.protocol.protoB\003IOP'))
_IOFLAGS.has_options = True
_IOFLAGS._options = _descriptor._ParseOptions(descriptor_pb2.EnumOptions(), _b('\020\001'))
# @@protoc_insertion_point(module_scope)



from typing import Union

from . import RequestMsg, ReplyMsg, Message, SimpleMessage
from hedgehog.protocol.errors import InvalidCommandError
from hedgehog.protocol.proto import motor_pb2
from hedgehog.protocol.proto.motor_pb2 import POWER, BRAKE, VELOCITY
from hedgehog.protocol.proto.subscription_pb2 import Subscription
from hedgehog.utils import protobuf


@RequestMsg.message(motor_pb2.MotorAction, 'motor_action')
class Action(SimpleMessage):
    def __init__(self, port: int, state: int, amount: int=0, reached_state: int=POWER, relative: int=None, absolute: int=None) -> None:
        if relative is not None and absolute is not None:
            raise InvalidCommandError("relative and absolute are mutually exclusive")
        if relative is None and absolute is None:
            if reached_state != 0:
                raise InvalidCommandError(
                    "reached_state must be kept at its default value for non-positional motor commands")
        else:
            if state == BRAKE:
                raise InvalidCommandError("state can't be BRAKE for positional motor commands")
            if amount <= 0:
                raise InvalidCommandError("velocity/power must be positive for positional motor commands")
        self.port = port
        self.state = state
        self.amount = amount
        self.reached_state = reached_state
        self.relative = relative
        self.absolute = absolute

    @classmethod
    def _parse(cls, msg: motor_pb2.MotorAction) -> 'Action':
        port = msg.port
        state = msg.state
        amount = msg.amount
        reached_state = msg.reached_state
        relative = msg.relative if msg.HasField('relative') else None
        absolute = msg.absolute if msg.HasField('absolute') else None
        return cls(port, state, amount, reached_state, relative, absolute)

    def _serialize(self, msg: motor_pb2.MotorAction) -> None:
        msg.port = self.port
        msg.state = self.state
        msg.amount = self.amount
        msg.reached_state = self.reached_state
        if self.relative is not None:
            msg.relative = self.relative
        if self.absolute is not None:
            msg.absolute = self.absolute


@RequestMsg.message(motor_pb2.MotorSetPositionAction, 'motor_set_position_action')
class SetPositionAction(SimpleMessage):
    def __init__(self, port: int, position: int) -> None:
        self.port = port
        self.position = position

    @classmethod
    def _parse(cls, msg: motor_pb2.MotorSetPositionAction) -> 'SetPositionAction':
        port = msg.port
        position = msg.position
        return cls(port, position)

    def _serialize(self, msg: motor_pb2.MotorSetPositionAction) -> None:
        msg.port = self.port
        msg.position = self.position


@protobuf.message(motor_pb2.MotorCommandMessage, 'motor_command_message', fields=('port',))
class CommandRequest(Message):
    def __init__(self, port: int) -> None:
        self.port = port

    def _serialize(self, msg: motor_pb2.MotorCommandMessage) -> None:
        msg.port = self.port


@protobuf.message(motor_pb2.MotorCommandMessage, 'motor_command_message', fields=('port', 'subscription'))
class CommandSubscribe(Message):
    def __init__(self, port: int, subscription: Subscription) -> None:
        self.port = port
        self.subscription = subscription

    def _serialize(self, msg: motor_pb2.MotorCommandMessage) -> None:
        msg.port = self.port
        msg.subscription.CopyFrom(self.subscription)


@RequestMsg.parser('motor_command_message')
def _parse_command_request(msg: motor_pb2.MotorCommandMessage) -> Union[CommandRequest, CommandSubscribe]:
    port = msg.port
    subscription = msg.subscription if msg.HasField('subscription') else None
    if subscription is None:
        return CommandRequest(port)
    else:
        return CommandSubscribe(port, subscription)


@protobuf.message(motor_pb2.MotorCommandMessage, 'motor_command_message', fields=('port', 'state', 'amount'))
class CommandReply(Message):
    def __init__(self, port: int, state: int, amount: int) -> None:
        self.port = port
        self.state = state
        self.amount = amount

    def _serialize(self, msg: motor_pb2.MotorCommandMessage) -> None:
        msg.port = self.port
        msg.state = self.state
        msg.amount = self.amount


@protobuf.message(motor_pb2.MotorCommandMessage, 'motor_command_message')
class CommandUpdate(Message):
    is_async = True

    def __init__(self, port: int, state: int, amount: int, subscription: Subscription) -> None:
        self.port = port
        self.state = state
        self.amount = amount
        self.subscription = subscription

    def _serialize(self, msg: motor_pb2.MotorCommandMessage) -> None:
        msg.port = self.port
        msg.state = self.state
        msg.amount = self.amount
        msg.subscription.CopyFrom(self.subscription)


@ReplyMsg.parser('motor_command_message')
def _parse_command_reply(msg: motor_pb2.MotorCommandMessage) -> Union[CommandReply, CommandUpdate]:
    port = msg.port
    state = msg.state
    amount = msg.amount
    subscription = msg.subscription if msg.HasField('subscription') else None
    if subscription is None:
        return CommandReply(port, state, amount)
    else:
        return CommandUpdate(port, state, amount, subscription)


@protobuf.message(motor_pb2.MotorStateMessage, 'motor_state_message', fields=('port',))
class StateRequest(Message):
    def __init__(self, port: int) -> None:
        self.port = port

    def _serialize(self, msg: motor_pb2.MotorStateMessage) -> None:
        msg.port = self.port


@protobuf.message(motor_pb2.MotorStateMessage, 'motor_state_message', fields=('port', 'subscription'))
class StateSubscribe(Message):
    def __init__(self, port: int, subscription: Subscription) -> None:
        self.port = port
        self.subscription = subscription

    def _serialize(self, msg: motor_pb2.MotorStateMessage) -> None:
        msg.port = self.port
        msg.subscription.CopyFrom(self.subscription)


@RequestMsg.parser('motor_state_message')
def _parse_state_request(msg: motor_pb2.MotorStateMessage) -> Union[StateRequest, StateSubscribe]:
    port = msg.port
    subscription = msg.subscription if msg.HasField('subscription') else None
    if subscription is None:
        return StateRequest(port)
    else:
        return StateSubscribe(port, subscription)


@protobuf.message(motor_pb2.MotorStateMessage, 'motor_state_message', fields=('port', 'velocity', 'position'))
class StateReply(Message):
    def __init__(self, port: int, velocity: int, position: int) -> None:
        self.port = port
        self.velocity = velocity
        self.position = position

    def _serialize(self, msg: motor_pb2.MotorStateMessage) -> None:
        msg.port = self.port
        msg.velocity = self.velocity
        msg.position = self.position


@protobuf.message(motor_pb2.MotorStateMessage, 'motor_state_message')
class StateUpdate(Message):
    is_async = True

    def __init__(self, port: int, velocity: int, position: int, subscription: Subscription) -> None:
        self.port = port
        self.velocity = velocity
        self.position = position
        self.subscription = subscription

    def _serialize(self, msg: motor_pb2.MotorStateMessage) -> None:
        msg.port = self.port
        msg.velocity = self.velocity
        msg.position = self.position
        msg.subscription.CopyFrom(self.subscription)


@ReplyMsg.parser('motor_state_message')
def _parse_state_reply(msg: motor_pb2.MotorStateMessage) -> Union[StateReply, StateUpdate]:
    port = msg.port
    velocity = msg.velocity
    position = msg.position
    subscription = msg.subscription if msg.HasField('subscription') else None
    if subscription is None:
        return StateReply(port, velocity, position)
    else:
        return StateUpdate(port, velocity, position, subscription)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: hedgehog/protocol/proto/subscription.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='hedgehog/protocol/proto/subscription.proto',
  package='hedgehog.protocol.proto',
  syntax='proto3',
  serialized_pb=_b('\n*hedgehog/protocol/proto/subscription.proto\x12\x17hedgehog.protocol.proto\"y\n\x0cSubscription\x12\x11\n\tsubscribe\x18\x01 \x01(\x08\x12\x0f\n\x07timeout\x18\x02 \x01(\r\x12\x1b\n\x13granularity_timeout\x18\x03 \x01(\r\x12\x19\n\x0fint_granularity\x18\x04 \x01(\rH\x00\x42\r\n\x0bgranularityB0\n\x1f\x61t.pria.hedgehog.protocol.protoB\rSubscriptionPb\x06proto3')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_SUBSCRIPTION = _descriptor.Descriptor(
  name='Subscription',
  full_name='hedgehog.protocol.proto.Subscription',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='subscribe', full_name='hedgehog.protocol.proto.Subscription.subscribe', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='timeout', full_name='hedgehog.protocol.proto.Subscription.timeout', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='granularity_timeout', full_name='hedgehog.protocol.proto.Subscription.granularity_timeout', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='int_granularity', full_name='hedgehog.protocol.proto.Subscription.int_granularity', index=3,
      number=4, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='granularity', full_name='hedgehog.protocol.proto.Subscription.granularity',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=71,
  serialized_end=192,
)

_SUBSCRIPTION.oneofs_by_name['granularity'].fields.append(
  _SUBSCRIPTION.fields_by_name['int_granularity'])
_SUBSCRIPTION.fields_by_name['int_granularity'].containing_oneof = _SUBSCRIPTION.oneofs_by_name['granularity']
DESCRIPTOR.message_types_by_name['Subscription'] = _SUBSCRIPTION

Subscription = _reflection.GeneratedProtocolMessageType('Subscription', (_message.Message,), dict(
  DESCRIPTOR = _SUBSCRIPTION,
  __module__ = 'hedgehog.protocol.proto.subscription_pb2'
  # @@protoc_insertion_point(class_scope:hedgehog.protocol.proto.Subscription)
  ))
_sym_db.RegisterMessage(Subscription)


DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(descriptor_pb2.FileOptions(), _b('\n\037at.pria.hedgehog.protocol.protoB\rSubscriptionP'))
# @@protoc_insertion_point(module_scope)
